package VR.alg

import java.io.{File, PrintWriter}

import VR.data.DataClass
import VR.prob.ProblemBase

import scala.util.Random

/**
  * Created by YY on 2016/1/7.
  */
class SAG[T<: DataClass](p:ProblemBase[T]) extends AlgorithmBase(p) {

  def shrink_val(num:Int):Int= num



  override def train(data:T, lambda: Double, sigma : Double, eta: Double,  tot_batch:Int , outputFile: Option[String]): Unit = {
    val (x_dat,y_dat,n,d)=data.get_data()
    w=new Array[Double](d)
    val shrink:Int=shrink_val(n)
    val cBlock=new Array[Array[Double]](n)
    val belongTo:Array[Int]=new Array[Int](n)
    for (i<-y_dat.indices){
      belongTo(i)=i
      cBlock(i)=prob.gradient_cal(w,data,i)
    }

    val save_tmp: Array[Double]=new Array[Double](1)

    val utilde:Array[Double]=new Array[Double](d)
    val thres:Double =math.abs(eta*sigma)

    val timeline:Array[timeLine]=new Array[timeLine](d)
    val fout:Option[PrintWriter]=outputFile match {
      case None => None
      case Some(st)=>Some(new PrintWriter(new File(st)))}
    var totGrad:Double=0
    var lastPrint:Double=0
    //below, compute the actual full gradient, utilde is not sparse.
    prob.fullGrad(data,w, utilde)
    totGrad += n
    for (i<- timeline.indices) timeline(i)=new timeLine(-1)
    var counter:Int= -1
    var get_out:Boolean=false

    val r_seed=new Random(1)

    while (!get_out)
    {
      counter+=1
      val randomRow:Int=r_seed.nextInt(n)
      val i:Int=belongTo(randomRow)
      val Bi:Array[Double]=cBlock(i)
      // all coordinates related to j are updated.
      for (j<- x_dat(randomRow))
        w(j._1) = update_w(w(j._1),timeline(j._1), counter, save_tmp,0, thres, 1.0-eta*lambda, eta*utilde(j._1))
      //compute g1 and g2 here.
      //val t1:Double=inner_sparse(w,x_dat(randomRow))-y_dat(randomRow)
      val t1:Array[Double]=prob.gradient_cal(w,data,randomRow)

      // - eta(utilde + \nabla_ f -\nabla_i f')
      for (j<-x_dat(randomRow).indices){
        // w - eta( utilde + g1-g2)
        val where=x_dat(randomRow)(j)._1
        w(where) -= eta*((t1(j) - Bi(j))/shrink + utilde(where))
      }
      // update utilde
      for (j<-x_dat(randomRow).indices)
        utilde(x_dat(randomRow)(j)._1) += (t1(j)-Bi(j))* 1.0 /n
      //proximal updates
      for (j<-x_dat(randomRow)){
        if (sigma>1e-10)
        {
          if (w(j._1) > thres)
            w(j._1) -= thres
          else if (w(j._1) < -thres)
            w(j._1) += thres
          else w(j._1)=0
        }
        else
          w(j._1) *= 1.0-eta*lambda
        timeline(j._1).last_touch=counter
      }
      cBlock(i)=t1
      totGrad+=1

      if (totGrad > lastPrint + n / sampleFreq)
      {
        for (j<-w.indices)
          w(j)=update_w(w(j),timeline(j), counter+1, save_tmp, 0, thres, 1.0-eta*lambda, eta*utilde(j))
        val curObj:Double=prob.full_obj_cal(data,w,lambda,sigma)
        fout match {
          case Some(a:PrintWriter)=>
            lastPrint=printToFile(curObj,totGrad,n,a)
          case None =>
        }
      }

      /*
      for (j<-w.indices){
        w(j)=update_w(w(j),timeline(j), t+1, save, j, eta*sigma, 1.0-eta*lambda, eta*utilde(j))
        timeline(j).last_touch=t
      }
      */
      if (counter%n==0){
        for (j<-w.indices)
          w(j)=update_w(w(j),timeline(j), counter+1, save_tmp, 0, thres, 1.0-eta*lambda, eta*utilde(j))
        val curObj:Double=prob.full_obj_cal(data,w,lambda,sigma)
        println(f"$curObj%.10f  eta=$eta%.5f  w0=${w(0)}%.5f  w1=${w(1)}%.5f  w2=${w(2)}%.5f")

        if (totGrad/tot_batch>=n)
          get_out=true
      }
    }

    fout match {
      case Some(a:PrintWriter) => a.close()
      case None =>
    }
  }

}
