package VR.alg

import java.io.{File, PrintWriter}

import VR.data.DataClass
import VR.prob.ProblemBase

import scala.util.Random

/**
  * Created by YY on 2016/1/8.
  */
class SDCA[T<:DataClass](p:ProblemBase[T],reg_lambda:Double) extends AlgorithmBase(p){
  val rlambda:Double=reg_lambda


  override def train(data: T, lambda: Double, sigma: Double, eta_no_use: Double, tot_batch: Int, outputFile: Option[String]): Unit = {
    val (x_dat,_,n,d)=data.get_data()
    val wt=Array(new Array[Double](d), new Array[Double](d))
    val vt=Array(new Array[Double](d), new Array[Double](d))

    val using_lambda:Double=if (lambda==0) rlambda else lambda
    val thres:Double =math.abs(sigma/using_lambda)

    val fout:Option[PrintWriter]=outputFile match {
      case None => None
      case Some(st)=>Some(new PrintWriter(new File(st)))}
    var lastPrint:Double=0
    val timeline:Array[Array[timeLine]]=Array(new Array[timeLine](d), new Array[timeLine](d))
    for (cur_t<-timeline)  for (i<- cur_t.indices) cur_t(i)=new timeLine(-1)
    val save:Array[Array[Double]]=Array(new Array[Double](d), new Array[Double](d))
    val alpha:Array[Array[Double]]=Array(new Array[Double](n), new Array[Double](n))
    val output:Array[Double]=new Array[Double](d)

    val r_seed:Array[Random]=Array(new Random(1), new Random(1))

    def get_prox(v:Double):Double={
      if (sigma>1e-10)
      {
        if (v > thres) v - thres
        else
          if (v < -thres) v + thres
          else 0
      } else v
    }

    for (counter <- Range(0, tot_batch*n)) {
      for (iter<-Range(0,2))
        if (counter%(iter+1)==0){
          val i:Int=r_seed(iter).nextInt(n)
          //val i:Int=(counter/(iter+1))%n
          w=wt(iter)
          val timeline_t=timeline(iter)
          val save_t=save(iter)
          val alpha_t=alpha(iter)
          val v=vt(iter)
          for (j<- x_dat(i))
            w(j._1)= get_prox(v(j._1))
          var middle:Double=prob.binary_alpha_cal(alpha_t(i), w,data, i, using_lambda, n)

          alpha_t(i) += middle
          for (j<-x_dat(i)){

            save_t(j._1) += get_prox(v(j._1)) * (counter/(iter+1)-timeline_t(j._1).last_touch)

            timeline_t(j._1).last_touch=counter/(iter+1)

            v(j._1) += 1.0/using_lambda/n*middle*j._2
          }
      }

      def compute_obj():Double={
        for (iter<-Range(0,2)){
          w=wt(iter)
          val v=vt(iter)
          val save_t=save(iter)
          val timeline_t = timeline(iter)
          for (j<-w.indices){
            save_t(j) += get_prox(v(j)) * (counter/(iter+1)-timeline_t(j).last_touch)
            timeline_t(j).last_touch=counter/(iter+1)
          }
        }
        for (j<-w.indices) output(j)=(save(0)(j)-save(1)(j))/(counter/2)
        prob.full_obj_cal(data,output,lambda,sigma)
      }
      if ((counter   > lastPrint + n / sampleFreq) && (counter%2==0)){
        val curObj:Double=compute_obj()
        fout match {
          case Some(a:PrintWriter)=>
            lastPrint=printToFile(curObj,counter,n,a)
          case None =>
        }
      }
      if ((counter%n<2) && (counter%2==0) && (counter>n/4)){
        val curObj:Double=compute_obj()
        println(f"$curObj%.10f  w0=${w(0)}%.5f  w1=${w(1)}%.5f  w2=${w(2)}%.5f")
      }
    }

    fout match {
      case Some(a:PrintWriter) => a.close()
      case None =>
    }

  }
}
