package VR.data

import VR._

import scala.util.Random

/**
  * Created by YY on 2016/1/6.
  */
object DataProcessor {
  def get_non_zero_and_avg_len_and_normalize(x:X_type):(Int, Double)={
    var cumulativeLen:Double=0
    var non_zero:Int=0
    for (item<-x){
      var sqrCurLen:Double=0
      non_zero+=item.length
      for (iter<-item)
        sqrCurLen+= sqr(iter._2)
      cumulativeLen+=math.sqrt(sqrCurLen)
    }
    val avgLen=cumulativeLen/x.length
    for (item<- x)
      for (i <- item.indices) {
        val tup = item(i)
        item(i) = tup.copy(_2 = tup._2 / avgLen)
      }
    (non_zero, cumulativeLen/x.length)
  }
  def set_all_norm_one(x:X_type):Unit={
    for (item<-x){
      var cumulativeLen:Double=0
      for (entry<-item)
        cumulativeLen+= VR.sqr(entry._2)
      cumulativeLen=math.sqrt(cumulativeLen)
      for (i<-item.indices){
        val tup=item(i)
        item(i)=tup.copy(_2=tup._2/cumulativeLen)
      }
    }
  }
  def random_gen(n:Int,d:Int):Data_all_content_type={
    val r_seed=new Random(1)
    val x:X_type=new Array[Single_x_type](n)
    val y:Y_type=new Array[Double](n)
    for (i<-Range(0,n)){
      x(i)=new Array[Sparse_type](d)
      for (j<-Range(0,d))
        if (r_seed.nextBoolean())
          x(i)(j)=(j,r_seed.nextDouble())
        else
          x(i)(j)=(j,-1*r_seed.nextDouble())
    }
    val (non_zero,avgLen)=get_non_zero_and_avg_len_and_normalize(x)
    (x,y,n,d,non_zero,avgLen)
  }
  def random_split(data: DataClass) : (DataClass, DataClass) = {
    val (x_dat,y_dat,n,d,non_zero,avg_len)=data.get_contents()
    val partition_frac:Double=0.2
    val r_seed=new Random(1)
    val split_x=Array(new X_type(n),new X_type(n))
    val split_y=Array(new Y_type(n),new Y_type(n))
    val count=Array(0,0)
    for (i <- y_dat.indices){
      val belong= if (r_seed.nextDouble>partition_frac) 0 else 1
      split_y(belong)(count(belong))=y_dat(i)
      split_x(belong)(count(belong))=x_dat(i)
      count(belong)+=1
    }
    (data.create_new_instance(split_x(0),split_y(0),split_x(0).length,d,non_zero,avg_len), data.create_new_instance(split_x(1), split_y(1),split_x(1).length,d,non_zero,avg_len))
  }
  def clusterReader(file_loc:String, content:Data_all_content_type): Array[Int]={
    val (_,_,n,_,_,_)=content
    val belongTo=new Array[Int](n)
    if (file_loc=="")
      return belongTo
    println("Opening file : "+file_loc)
    val cluster_source=io.Source.fromFile(file_loc)
    val cluster_info=cluster_source.getLines().next()
    val elements=cluster_info.split(" ")
    for (ele<-Range(0,elements.length-1))
      belongTo(ele)=elements(ele).toInt
    belongTo
  }

  def dataReader(file_loc:String, helper_loc:String, is_logistic:Boolean, reserve_col_zero:Boolean=true): Data_all_content_type = {
    println("Opening file : "+file_loc+" \nwith helper file : "+helper_loc)
    val helper_source=io.Source.fromFile(helper_loc)
    var i:Int=0
    var n:Int=0
    var logo:String=""
    for (line<-helper_source.getLines())
      if (i<2){
        if (i==0) n=line.toInt
        else logo=line.toString
        i+=1
      }
    helper_source.close()
    val data_source=io.Source.fromFile(file_loc)
    val lines = data_source.getLines()

    val dat_x:X_type=new X_type(n)
    val dat_y:Y_type=new Y_type(n)
    var d=0

    var line_n=0
    println("n=%d".format(n))
    println("Reading data...")
    for (line <- lines){
      val elements=line.split(" ")
      var i=0
      val len=elements.length
      val shift:Int= if (reserve_col_zero) 0 else -1
      val item:Single_x_type=new Single_x_type(len+shift)
      var sqrLen:Double=0
      dat_x(line_n)=item
      for (ele<- elements) {
        if (i==0){
          if (ele.equals(logo)){
            dat_y(line_n)= 1
          }
          else
            dat_y(line_n)= -1
          if (reserve_col_zero){
            item(i)=(0,1)
            sqrLen+= sqr(item(i)._2)
          }
        }
        else{
          val parts=ele.split(":")
          item(i+shift)=(parts(0).toInt, parts(1).toDouble)
          sqrLen+= sqr(parts(1).toDouble)
          if (d<=parts(0).toInt)
            d=parts(0).toInt+1
          if ((parts(0).toInt==0) && reserve_col_zero)
            println("Error: I didn't expect feature at location 0. That's reserved for constant 1. ")
        }
        i=i+1
      }
      if (is_logistic){
        for (i<-item.indices)
          item(i)=item(i).copy(_2=item(i)._2*dat_y(line_n))
        dat_y(line_n)=1
      }
      line_n+=1
    }
    println("Scaling..")
    data_source.close()
    val (avgLen,tot_non_zero)=get_non_zero_and_avg_len_and_normalize(dat_x)
    println("tot line=",n)
    println("Done Scaling..!")

    (dat_x,dat_y,n, d,avgLen, tot_non_zero)
  }
}
