package VR.data

import java.io.{File, PrintWriter, OutputStreamWriter}

import VR._

import scala.util.Random

/**
  * Created by YY on 2016/1/23.
  */
class PCA_Data(content:Data_all_content_type, diagonal_type:Int, delta:Double) extends DataClass(content) {
  DataProcessor.set_all_norm_one(x)
  val full_matrix:Matrix_type=compute_full_matrix()
  val matrix_D:Array[Row_type]=get_Diagonal_D(diagonal_type,delta, n, d)
  val random_b:Array[Sparse_type]=get_random_b(d)
  def this(gen_n:Int, gen_d:Int)=this(DataProcessor.random_gen(gen_n,gen_d), 0,0)
  def this(diagonal_type:Int, delta:Double, file_loc:String, helper_loc:String)=this(DataProcessor.dataReader(file_loc, helper_loc, is_logistic = false, reserve_col_zero = false),diagonal_type,delta)

  def get_random_b(d:Int): Array[Sparse_type]={
    val r_seed=new Random(7)
    val output:Array[Sparse_type]=new Array[Sparse_type](d)
    for (i<- Range(0,d))
      output(i)=(i,r_seed.nextDouble()/100)
    output
  }

  def compute_full_matrix(): Matrix_type={
    val matrix_A:Matrix_type=new Matrix_type(d)
    for (i<-Range(0,d))
      matrix_A(i)=new Row_type(d)
    for (t<- Range(0, n)){
      for (i<-Range (0, d))
        for (j<-Range(0,d))
          matrix_A(i)(j) += x(t)(i)._2* x(t)(j)._2
    }
    for (i<-Range(0,d))
      for (j<-Range(0,d))
        matrix_A(i)(j)/=n
    matrix_A
  }
  def print_matrix(print_to_file:String="", m:Matrix_type):Unit={
      val f_out=if (print_to_file=="") new OutputStreamWriter(System.out)
                else new PrintWriter(new File(print_to_file))
      for (row<- m.indices) {
        for (col<- m(row).indices)
          f_out.write(m(row)(col).toString+" ")
        f_out.write("\n")
      }
      f_out.close()
  }
  def print_norms():Unit={
    for (item<-x){
      var cumulativeLen:Double=0
      for (entry<-item)
        cumulativeLen+= VR.sqr(entry._2)
      cumulativeLen=math.sqrt(cumulativeLen)
      println(cumulativeLen)
    }
  }
  def verify_sum_d():Unit={
    for (i<-Range(0, d)){
      var sum:Double=0
      for (j<-Range (0,n))
        sum+= matrix_D(j)(i)
      if (math.abs(sum)>1e-10){
        println("Error: sum not equal to 0")
        assert(false)
      }
    }
    println("-------------  Passed verification!")
  }
  def get_Diagonal_D(diagonal_type:Int, delta:Double, n :Int, d:Int):Array[Row_type]={
    if (diagonal_type==0)
      return new Array[Row_type](1)
    println("type="+diagonal_type+" n="+n+" d="+d)
    val r_seed=new Random(diagonal_type+2)
    val matrix_D:Array[Row_type]=new Array[Row_type](n)
    for (i<-matrix_D.indices) matrix_D(i)=new Array[Double](d)
    val count_on_j:List[Array[Int]]=List(new Array[Int](d), new Array[Int](d))
    if (diagonal_type==1){
      for (i <- Range(0, n)){
        for (j <- Range(0,d)){
          val position:Int=
            if (((count_on_j.head(j) < n/2) && r_seed.nextBoolean()) || (count_on_j(1)(j)==n/2)) 0 else 1
          count_on_j(position)(j) += 1
          matrix_D(i)(j)=if (position==0) delta else -delta
        }
      }
      matrix_D
    } else{
      for (i <- Range(0, d)){
        val where_is_big:Int=r_seed.nextInt(n)
        for (j <- Range(0,n))
          matrix_D(j)(i)=delta/(n-1)
        matrix_D(where_is_big)(i)= -delta
      }
      matrix_D
    }
  }
}
